""":demand: F1.8"""

from collections.abc import Iterable
from typing import TypeVar
from .List import List
from .Node import Node

T = TypeVar('T')


class LinkedList(List):
    """
    A linked list that consists of a set of sequentially linked Node.
    """

    # head -> ... -> ... -> ... -> sentinel()
    def __init__(self, it: Iterable[T] = None) -> None:
        self.__head: Node = Node.sentinel()
        self.__length: int = 0
        if it is not None:
            for item in it[::-1]:
                self.insert_head(item)

    @property
    def head(self) -> Node:
        """
        Returns the head value of the list.

        :return: TwoWayNode
        """
        return self.__head.value

    @property
    def length(self) -> int:
        """
        Returns the length of the list.

        :return: int
        """
        return self.__length

    def is_empty(self) -> bool:
        """
        Checks if the list is empty.

        :return: bool
        """
        return self.length == 0

    def get(self, index: int) -> T:
        """
        Returns the element at the `index` position.

        :param index: int
        :return: Object
        """
        return self._get_node(index).value

    def _get_node(self, index: int) -> Node:
        """
        Returns the node at the `index` position.

        :param index: int
        :return: TwoWayNode
        """
        if index < 0 or index >= self.length:
            raise IndexError('Index out of range')
        i: int = 0
        current_node = self.__head
        while i < index:
            current_node = current_node.next
            i += 1
        return current_node

    def set(self, index: int, e: T) -> None:
        """
        Sets the element at the `index` position as `e`.

        :param index: int
        :param e: Object
        :return: None
        """
        node: T = self._get_node(index)
        node.value = e

    def insert_head(self, e: T) -> None:
        """
        Inserts the element at the head of the list.

        :param e: Object
        :return: None
        """
        new_node = Node(e, self.__head)
        self.__head = new_node
        self.__length += 1

    def insert_after(self, index: int, e: T) -> None:
        """
        Inserts the element after the given `index`.

        :param index: int
        :param e: Object
        :return: None
        """
        if 0 < index or index >= self.__length:
            raise IndexError('Index out of range')
        new_node = Node(e)
        current_node = self._get_node(index)
        new_node.next = current_node.next
        current_node.next = new_node
        self.__length += 1

    def insert(self, index: int, e: T) -> None:
        if index == 0:
            return self.insert_head(e)
        self.insert_after(index - 1, e)

    def delete(self, index: int) -> None:
        """
        Deletes the element at the given `index`.

        :param index: int
        :return: None
        """
        if index < 0 or index >= self.__length:
            raise IndexError('Index out of range')
        self.__length -= 1
        if index == 0:
            self.__head = self.__head.next
            return
        current_node = self._get_node(index - 1)
        current_node.next = current_node.next.next

    def index(self, v: T) -> int:
        """
        Returns the index of the element `e` in the list. Raises ValueError if `e` is not in the list.

        :param v: Object
        :return: int
        """
        i: int = 0
        current_node = self.__head
        while not current_node.is_sentinel() and current_node.value != v:
            current_node = current_node.next
            i += 1
        if current_node.is_sentinel():
            raise ValueError(f'No such element in the list :{v}')
        return i

    def __str__(self) -> str:
        if self.is_empty():
            return "[]"
        current_node = self.__head
        string: str = "[{}".format(current_node.value)

        while current_node.has_next():
            current_node = current_node.next
            if current_node.has_next():
                string += "→{}".format(current_node.value)
        return string + "]"

    def __getitem__(self, item: int) -> Node:
        return self.get(item)

    def __eq__(self, other: T) -> bool:
        if not isinstance(other, LinkedList) or self.length != other.length:
            return False
        for i in range(self.length):
            if self[i] != other[i]:
                return False
        return True

    def __len__(self) -> int:
        return self.length

    def __iter__(self):
        for i in range(self.length):
            yield self[i]

    def __to_visu__(self) -> str:
        if self.is_empty():
            return ""
        current_node = self.__head
        string: str = "{}".format(current_node.value)

        while current_node.has_next():
            current_node = current_node.next
            if current_node.has_next():
                string += ",{}".format(current_node.value)
        return string


def LinkedList_length(list: LinkedList) -> int:
    return list.length


def LinkedList_get(list: LinkedList, index: int) -> T:
    return list.get(index)


def LinkedList_insert(list: LinkedList, index: int, e: T) -> None:
    list.insert(index, e)


def LinkedList_delete(list: LinkedList, index: int) -> None:
    list.delete(index)


def LinkedList_is_empty(list: LinkedList) -> bool:
    return list.is_empty()


def LinkedList_set(list: LinkedList, index: int, e: T) -> None:
    list.set(index, e)


def LinkedList_insert_head(list: LinkedList, e: T) -> None:
    list.insert_head(e)


def LinkedList_insert_after(list: LinkedList, index: int, e: T) -> None:
    list.insert_after(index, e)


def LinkedList__get_node(list: LinkedList, index: int):
    return list._get_node(index)
